import base64
import datetime
from threading import Thread
from typing import Callable, ClassVar, Optional

import pandas as pd
import requests

from naneos.logger import get_naneos_logger
from naneos.protobuf.protobuf import (
    create_combined_entry,
    create_proto_p1,
    create_proto_p2,
    create_proto_p2_pro,
    create_proto_p2_pro_garage,
)

logger = get_naneos_logger(__name__)


class NaneosUploadThread(Thread):
    URL: ClassVar[str] = "https://hg3zkburji.execute-api.eu-central-1.amazonaws.com/dev/proto/v1"
    HEADERS: ClassVar[dict] = {"Content-Type": "application/json", "Accept": "application/json"}

    def __init__(
        self, data: list[tuple[int, str, pd.DataFrame]], callback: Optional[Callable[[bool], None]]
    ) -> None:
        """Adding the data that should be uploaded to the database.

        Args:
            data (list[tuple[int, pd.DataFrame]]):
                int: serial number
                str: device type
                pd.DataFrame: data to upload
            callback (Optional[Callable[[bool], None]]): Callback function that is called after upload.
        """
        super().__init__()
        self.data = data
        self._callback = callback

    def run(self) -> None:
        try:
            ret = self.upload(self.data)

            if self._callback:
                if ret.status_code == 200:
                    self._callback(True)
                else:
                    self._callback(False)
        except Exception as e:
            logger.error(f"Error in upload: {e}")
            if self._callback:
                self._callback(True)  # delete data because it was corrupted

    @staticmethod
    def get_body(upload_string: str) -> str:
        return f"""
            {{
                "gateway": "python_webhook",
                "data": "{upload_string}",
                "published_at": "{datetime.datetime.now().isoformat()}"
            }}
            """

    @classmethod
    def upload(cls, data: list[tuple[int, str, pd.DataFrame]]) -> requests.Response:
        abs_time = int(datetime.datetime.now().timestamp())
        devices = []

        for device in data:
            sn = device[0]
            device_type = device[1]
            df = device[2]

            if device_type == "P1":
                devices.append(create_proto_p1(sn, abs_time, df))
            elif device_type == "P2":
                devices.append(create_proto_p2(sn, abs_time, df))
            elif device_type == "P2_Pro":
                devices.append(create_proto_p2_pro(sn, abs_time, df))
            elif device_type == "P2_Pro_Garage":
                devices.append(create_proto_p2_pro_garage(sn, abs_time, df))
            else:
                raise ValueError(f"Unknown device type: {device_type}")

        combined_entry = create_combined_entry(devices=devices, abs_timestamp=abs_time)

        proto_str = combined_entry.SerializeToString()
        proto_str_base64 = base64.b64encode(proto_str).decode()

        body = cls.get_body(proto_str_base64)
        r = requests.post(cls.URL, headers=cls.HEADERS, data=body, timeout=10)
        # print(f"Status code: {r.status_code} text={r.text}")
        return r


if __name__ == "__main__":
    df_p2: pd.DataFrame = pd.read_pickle(
        "/Users/huegi/gitlocal/naneos/naneos-devices/tests/p2_test_data.pkl"
    )
    df_p2_pro: pd.DataFrame = pd.read_pickle(
        "/Users/huegi/gitlocal/naneos/naneos-devices/tests/p2_pro_test_data.pkl"
    )

    def callback(ret: bool) -> None:
        print(f"Callback: {ret}")

    data = [(int(777), str("P2_Pro"), df_p2_pro), (int(666), str("P2"), df_p2)]  # noqa: UP018
    # NaneosUploadThread.upload(data)

    t = NaneosUploadThread(data, callback)
    t.start()
    t.join()



import asyncio
from datetime import datetime, timezone
from queue import Queue
import subprocess
import sys
from threading import Event, Lock, Thread
import time

from bleak import BleakScanner
import requests

# TODO: refactoring / maybe thread is not needed


class Partector2Ble(Thread):
    def __init__(self) -> None:
        self.__init_thread()
        self.__init_data_structures()

        self.start()

    def __init_thread(self):
        Thread.__init__(self)
        self.name = "ble-partector2"
        self.STOP_EVENT = Event()

    def __init_data_structures(self):
        self.__next_scanning_time = time.time() - 1.0  # starts directly
        self._scanning_data = Queue()
        self._results_dict = {}
        self._results_lock = Lock()

    def run(self):
        print("Starting Partector2 BLE thread")
        asyncio.run(self._main_async())
        print("Closed Partector2 BLE thread and event loop")

    def stop_scanning(self):
        self.STOP_EVENT.set()
        self.join()
        print("Joined Partector2 BLE thread")

    async def _main_async(self):
        await asyncio.gather(self.__scan_in_background(), self.__decode_beacon_data())

    async def __scan_in_background(self):
        while not self.STOP_EVENT.is_set():
            self.__wait_until_trigger()
            self.__update_next_scanning_time()

            # fetching ble readings + saving the timestamp
            tmp_devices = await BleakScanner.discover(timeout=0.8)  # TODO: check timeout
            date_time = datetime.now(tz=timezone.utc)

            for d in (x for x in tmp_devices if x.name == "P2"):
                self._scanning_data.put([date_time, d])

            # handles blueZ error on raspberry pi's (20.11.2021)
            if "linux" in sys.platform:
                self.__clean_bluez_cache()
        print("Stopped scanning")

    def __update_next_scanning_time(self):
        if abs(time.time() - self.__next_scanning_time) < 0.1:
            self.__next_scanning_time += 1.0
        else:
            self.__next_scanning_time = time.time() + 1

    def __wait_until_trigger(self):
        # sleep until the specified datetime
        dt = self.__next_scanning_time - time.time()

        if dt > 0.0 and dt < 1.0:
            time.sleep(dt)

    # method is linux (+ maybe macos) only
    def __clean_bluez_cache(self):
        # get str with all cached devices from shell
        cached_str = subprocess.check_output("bluetoothctl devices", shell=True).decode(
            sys.stdout.encoding
        )

        # str manipulating and removing the apropreate P2 mac-addresses
        cached_list = cached_str.splitlines()
        for line in (x for x in cached_list if "P2" in x):
            # line format: "Device XX:XX:XX:XX:XX:XX P2"
            line = line.replace("Device", "").replace("P2", "").replace(" ", "")
            subprocess.check_output(f"bluetoothctl -- remove {line}", shell=True)

    async def __decode_beacon_data(self):
        while not self.STOP_EVENT.is_set():
            scan_data = list(self._scanning_data.queue)
            self._scanning_data.queue.clear()

            for d in scan_data:
                timestamp = d[0]
                scan_result = d[1]

                beac_meta = scan_result.metadata["manufacturer_data"]
                beac_bytes = list(beac_meta.keys())[0].to_bytes(2, byteorder="little")
                beac_bytes += list(beac_meta.values())[0]

                # if first byte is no X there is something wrong
                if chr(beac_bytes[0]) == "X" and len(beac_bytes) == 22:
                    measurement = self.__parse_mesurment_data(timestamp, beac_bytes)
                    self.__add_measurement_to_results(measurement)

            await asyncio.sleep(3)
        print("Stopped decoding")

    def __parse_mesurment_data(self, timestamp, b: bytes) -> dict:
        """Returns a dict with the serial number as key."""
        serial_number = int(int(b[15]) + (int(b[16]) << 8))
        measurement = {
            "dateTime": timestamp,
            "LDSA": (int(b[1]) + (int(b[2]) << 8) + (int(b[3]) << 16)) / 100.0,
            "diameter": float(int(b[4]) + (int(b[5]) << 8)),
            "number": float(int(b[6]) + (int(b[7]) << 8) + (int(b[8]) << 16)),
            "T": float(int(b[9])),
            "RHcorr": float(int(b[10])),
            "device_status": int(b[11])
            + (int(b[11]) << 8)
            + (((int(b[20]) >> 1) & 0b01111111) << 16),
            "batt_voltage": (int(b[13]) + (int(b[14]) << 8)) / 100.0,
            "particle_mass": (int(b[17]) + (int(b[18]) << 8) + (int(b[19]) << 16)) / 100.0,
        }

        return {serial_number: measurement}

    def __add_measurement_to_results(self, measurement: dict):
        sn, data = list(measurement.items())[0]

        with self._results_lock:
            if sn in list(self._results_dict.keys()):
                self._results_dict[sn].append(data)
            else:
                self._results_dict[sn] = [data]

    def get_and_clear_results(self):
        with self._results_lock:
            results = self._results_dict
            self._results_dict = {}
        return results


def test():
    ble_scanner = Partector2Ble()

    while True:
        try:
            time.sleep(5)
            print("---")
            data = ble_scanner.get_and_clear_results()
            print(data)

        except KeyboardInterrupt:
            break
        except Exception as excep:
            print(f"Excepion during download from P2: {excep}")

    ble_scanner.stop_scanning()
    print("stopped")


if __name__ == "__main__":
    test()



from queue import Queue
from threading import Thread
from typing import Any, Callable, Optional

import serial

from naneos.logger.custom_logger import get_naneos_logger
from naneos.partector.blueprints._data_structure import PARTECTOR2_DATA_STRUCTURE_V_LEGACY
from naneos.partector.blueprints._partector_blueprint import PartectorBluePrint
from naneos.serial_utils import list_serial_ports as ls_ports

logger = get_naneos_logger(__name__)


class ScanPartector(PartectorBluePrint):
    def __init__(
        self, serial_number: Optional[int] = None, port: Optional[str] = None, verb_freq: int = 1
    ) -> None:
        super().__init__(serial_number, port, verb_freq)

    def _init_serial_data_structure(self) -> None:
        self._data_structure = PARTECTOR2_DATA_STRUCTURE_V_LEGACY
        # not used for scanning

    def _serial_wrapper(self, func: Callable[[], Any]) -> Optional[Any]:
        """Wraps user func in try-except block. Forwards exceptions to the user."""
        if not self._connected:
            return None

        excep = "Was not able to fetch the serial number!"

        for _ in range(self.SERIAL_RETRIES):
            try:
                return func()
            except Exception as e:
                # ßlogger.error(f"SN{self._sn} Exception in _serial_wrapper: {e}")
                excep = f"SN{self._sn} Exception occured during user function call: {e}"

        raise Exception(excep)

    def _init_get_device_info(self) -> None:
        try:
            if self._sn is None:
                self._sn = self._get_serial_number_secure()
            self._fw = self.get_firmware_version()
            logger.debug(f"Connected to SN{self._sn} on {self._port}")
        except Exception:
            pass
            # logger.warning("Could not get device info!")

    def _set_verbose_freq(self, freq: int) -> None:
        """
        Set the frequency of the verbose output.

        :param int freq: Frequency of the verbose output in Hz. (0: off, 1: 1Hz, 2: 10Hz, 3: 100Hz)
        """

        if freq < 0 or freq > 3:
            raise ValueError("Frequency must be between 0 and 3!")

        self._write_line(f"X000{freq}!")

    def _check_connection(self) -> None:
        if isinstance(self._ser, serial.Serial) and self._ser.is_open:
            self._connected = True


def scan_for_serial_partector(serial_number: int, partector_version: str) -> Optional[str]:
    """Scans all possible ports using threads (fast)."""
    threads = []
    q_1 = Queue()
    q_2 = Queue()
    q_2_pro = Queue()
    q_2_pro_cs = Queue()

    [
        threads.append(Thread(target=__scan_port, args=(port, q_1, q_2, q_2_pro, q_2_pro_cs)))
        for port in ls_ports()
    ]
    [thread.start() for thread in threads]
    [thread.join() for thread in threads]

    q = Queue()
    if partector_version == "P1":
        q = q_1
    elif partector_version == "P2":
        q = q_2
    elif partector_version == "P2_Pro":
        q = q_2_pro
    elif partector_version == "P2proCS":
        q = q_2_pro_cs

    ports = {k: v for d in tuple(q.queue) for (k, v) in d.items()}
    logger.debug(f"Found ports: {ports}")

    if serial_number in ports:
        return ports[serial_number]

    return None


def scan_for_serial_partectors(sn_exclude: Optional[list] = None) -> dict:
    """Scans all possible ports using threads (fast)."""
    threads = []
    q_1 = Queue()
    q_2 = Queue()
    q_2_pro = Queue()
    q_2_pro_cs = Queue()

    if sn_exclude is None:
        sn_exclude = []

    [
        threads.append(Thread(target=__scan_port, args=(port, q_1, q_2, q_2_pro, q_2_pro_cs)))
        for port in ls_ports()
        if port not in sn_exclude
    ]
    [thread.start() for thread in threads]
    [thread.join() for thread in threads]

    p1 = {k: v for x in tuple(q_1.queue) for (k, v) in x.items()}
    p2 = {k: v for x in tuple(q_2.queue) for (k, v) in x.items()}
    p2_pro = {k: v for x in tuple(q_2_pro.queue) for (k, v) in x.items()}
    p2_pro_cs = {k: v for x in tuple(q_2_pro_cs.queue) for (k, v) in x.items()}

    return {"P1": p1, "P2": p2, "P2_Pro": p2_pro, "P2proCS": p2_pro_cs}


def __scan_port(port: str, q_1: Queue, q_2: Queue, q_2_pro: Queue, q_2_pro_cs: Queue) -> None:
    partector: Optional[ScanPartector] = None  # Initialize p2 with a default value of None

    try:
        partector = ScanPartector(port=port)

        if partector._sn is None:
            pass
        elif partector._sn < 1000:
            q_1.put({partector._sn: port})
        else:
            if partector._fw < 310:
                q_2.put({partector._sn: port})
            else:
                name: str = partector.write_line("name?")[1]
                if name == "P2":
                    q_2.put({partector._sn: port})
                elif name == "P2pro":
                    q_2_pro.put({partector._sn: port})
                elif name == "P2proCS":
                    q_2_pro_cs.put({partector._sn: port})

        partector.close(blocking=True)
    except Exception:
        if partector is not None:
            partector.close(blocking=True)


if __name__ == "__main__":
    print(scan_for_serial_partectors())



import base64
import datetime
import time
from typing import Optional

import pandas as pd

import naneos.protobuf.protoV1_pb2 as pbScheme


def create_combined_entry(
    devices: list[pbScheme.Device],
    abs_timestamp: int,
    gateway_points: Optional[list[pbScheme.GatewayPoint]] = None,
    position_points: Optional[list[pbScheme.PositionPoint]] = None,
    wind_points: Optional[list[pbScheme.WindPoint]] = None,
) -> pbScheme.CombinedData:
    combined = pbScheme.CombinedData()
    combined.abs_timestamp = abs_timestamp

    combined.devices.extend(devices)

    if gateway_points is not None:
        combined.gateway_points.extend(gateway_points)

    if position_points is not None:
        combined.position_points.extend(position_points)

    if wind_points is not None:
        combined.wind_points.extend(wind_points)

    return combined


def create_proto_p1(sn: int, abs_time: int, df: pd.DataFrame) -> pbScheme.Device:
    device = pbScheme.Device()
    device.type = 0
    device.serial_number = sn

    device_points = df.apply(_create_device_point, axis=1, abs_time=abs_time).to_list()  # type: ignore
    device_points = [x for x in device_points if x is not None]

    device.device_points.extend(device_points)

    return device


def create_proto_p2(sn: int, abs_time: int, df: pd.DataFrame) -> pbScheme.Device:
    device = pbScheme.Device()
    device.type = 0
    device.serial_number = sn

    device_points = df.apply(_create_device_point, axis=1, abs_time=abs_time).to_list()  # type: ignore
    device_points = [x for x in device_points if x is not None]

    device.device_points.extend(device_points)

    return device


def create_proto_p2_pro(sn: int, abs_time: int, df: pd.DataFrame) -> pbScheme.Device:
    device = pbScheme.Device()
    device.type = 0
    device.serial_number = sn

    device_points = df.apply(_create_device_point, axis=1, abs_time=abs_time).to_list()  # type: ignore
    device_points = [x for x in device_points if x is not None]

    device.device_points.extend(device_points)

    return device


def create_proto_p2_pro_garage(sn: int, abs_time: int, df: pd.DataFrame) -> pbScheme.Device:
    device = pbScheme.Device()
    device.type = 0
    device.serial_number = sn

    device_points = df.apply(_create_device_point, axis=1, abs_time=abs_time).to_list()  # type: ignore
    device_points = [x for x in device_points if x is not None]

    device.device_points.extend(device_points)

    return device


def _create_device_point(ser: pd.Series, abs_time: int) -> Optional[pbScheme.DevicePoint]:
    try:
        device_point = pbScheme.DevicePoint()

        # mandatory fields
        if isinstance(ser.name, int):
            timestamp = ser.name
        else:
            raise ValueError("Timestamp is not an int!")
        device_point.timestamp = abs_time - timestamp
        device_point.device_status = int(ser["device_status"])

        # optional fields
        if "idiff_global" in ser:
            idiff_tmp = ser["idiff_global"] if ser["idiff_global"] > 0 else 0
            device_point.diffusion_current = int(idiff_tmp * 100.0)
        if "ucor_global" in ser:
            device_point.corona_voltage = int(ser["ucor_global"])
        if "hiresADC1" in ser:
            pass
        if "hiresADC2" in ser:
            pass
        if "EM_amplitude1" in ser:
            pass
        if "EM_amplitude2" in ser:
            pass
        if "T" in ser:
            device_point.temperature = int(ser["T"])
        if "RHcorr" in ser:
            device_point.relative_humidity = int(ser["RHcorr"])
        if "deposition_voltage" in ser:
            device_point.deposition_voltage = int(ser["deposition_voltage"])
        if "batt_voltage" in ser:
            device_point.battery_voltage = int(ser["batt_voltage"] * 100.0)
        if "flow_from_dp" in ser:
            device_point.flow = int(ser["flow_from_dp"] * 1000.0)
        if "LDSA" in ser:
            device_point.ldsa = int(ser["LDSA"] * 100.0)
        if "diameter" in ser:
            device_point.average_particle_diameter = int(ser["diameter"])
        if "number" in ser:
            device_point.particle_number_concentration = int(ser["number"])
        if "dP" in ser:
            pass
        if "P_average" in ser:
            device_point.ambient_pressure = int(ser["P_average"] * 10.0)
        if "em_gain1" in ser:  # multiplicator ???
            device_point.electrometer_gain = int(ser["em_gain1"] * 100.0)
        if "em_gain2" in ser:
            device_point.electrometer_2_gain = int(ser["em_gain2"] * 100.0)

        # P2 Pro
        if "surface" in ser:
            device_point.surface = int(ser["surface"] * 100.0)
        if "particle_mass" in ser:
            device_point.particle_mass = int(ser["particle_mass"] * 100.0)
        if "sigma" in ser:
            device_point.sigma_size_dist = int(ser["sigma"] * 100.0)
        if "pump_current" in ser:
            device_point.pump_current = int(ser["pump_current"] * 100.0)
        if "pump_pwm" in ser:
            device_point.pump_pwm = int(ser["pump_pwm"])
        if "steps" in ser:
            device_point.steps_inversion = int(ser["steps"])
        if "current_0" in ser:
            device_point.current_dist_0 = int(ser["current_0"])
        if "current_1" in ser:
            device_point.current_dist_1 = int(ser["current_1"])
        if "current_2" in ser:
            device_point.current_dist_2 = int(ser["current_2"])
        if "current_3" in ser:
            device_point.current_dist_3 = int(ser["current_3"])
        if "current_4" in ser:
            device_point.current_dist_4 = int(ser["current_4"])
        if "particle_number_10nm" in ser:
            device_point.particle_number_10nm = int(ser["particle_number_10nm"])
        if "particle_number_16nm" in ser:
            device_point.particle_number_16nm = int(ser["particle_number_16nm"])
        if "particle_number_26nm" in ser:
            device_point.particle_number_26nm = int(ser["particle_number_26nm"])
        if "particle_number_43nm" in ser:
            device_point.particle_number_43nm = int(ser["particle_number_43nm"])
        if "particle_number_70nm" in ser:
            device_point.particle_number_70nm = int(ser["particle_number_70nm"])
        if "particle_number_114nm" in ser:
            device_point.particle_number_114nm = int(ser["particle_number_114nm"])
        if "particle_number_185nm" in ser:
            device_point.particle_number_185nm = int(ser["particle_number_185nm"])
        if "particle_number_300nm" in ser:
            device_point.particle_number_300nm = int(ser["particle_number_300nm"])

        # P2 Pro Garage
        if "cs_status" in ser:
            device_point.cs_status = int(ser["cs_status"])

    except Exception as e:
        print(f"Error in _create_device_Point: {e}")
        return None

    return device_point



class PartectorDefaults:
    """Default values for the Partector when connected via serial port."""

    SERIAL_RETRIES = 7
    SERIAL_TIMEOUT = 0.2
    SERIAL_TIMEOUT_INFO = SERIAL_TIMEOUT + 0.05
    SERIAL_BAUDRATE = 115200
    SERIAL_QUEUE_MAXSIZE = 200
    SERIAL_INFO_QUEUE_MAXSIZE = 20



from collections import deque
import time
from typing import Optional

from bleak import BleakClient
from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.scanner import AdvertisementData

from naneos.logger import LEVEL_DEBUG, LEVEL_INFO, get_naneos_logger
from naneos.partector.blueprints._data_structure import Partector2DataStructure

logger = get_naneos_logger(__name__, LEVEL_INFO)


class PartectorBleDevice:
    def __init__(self, serial_number: int) -> None:
        self.serial_number: int = serial_number
        self.ble_client: Optional[BleakClient] = None
        self._data_queue: deque[Partector2DataStructure] = deque(maxlen=100)

    def _add_old_format_data(self, data: AdvertisementData, serial_number: int) -> None:
        """Adds data from old format to the data queue"""
        adv, sn_compare = self.get_naneos_adv(data)

        if sn_compare != serial_number or adv is None:
            logger.warning(
                f"Serial number from advertisement data ({sn_compare}) does not match serial number from connection ({serial_number})"
            )
            return None

        decoded_data = self.decode_std_chareristic(bytearray(adv))
        decoded_data.unix_timestamp = int(time.time())

        if len(self._data_queue) == self._data_queue.maxlen:
            self._data_queue.popleft()
        self._data_queue.append(decoded_data)

        logger.debug(f"Added old format data from {self.serial_number}")

    def callback_std(self, characteristic: BleakGATTCharacteristic, data: bytearray) -> None:
        timestamp = int(time.time())
        decoded_data = self.decode_std_chareristic(data)
        decoded_data.unix_timestamp = timestamp
        if len(self._data_queue) == self._data_queue.maxlen:
            self._data_queue.popleft()
        self._data_queue.append(decoded_data)

        logger.debug(f"Callback std from {self.serial_number}")

    def callback_aux(self, characteristic: BleakGATTCharacteristic, data: bytearray) -> None:
        decoded_data = self.decode_aux_characteristic(data)
        if self._data_queue:
            for field in self._data_queue[-1].__dataclass_fields__:
                if getattr(decoded_data, field) is not None:
                    setattr(self._data_queue[-1], field, getattr(decoded_data, field))

        logger.debug(f"Callback aux from {self.serial_number}")

    def callback_size_dist(self, characteristic: BleakGATTCharacteristic, data: bytearray) -> None:
        decoded_data = self.decode_size_dist_characteristic(data)
        if self._data_queue:
            for field in self._data_queue[-1].__dataclass_fields__:
                if getattr(decoded_data, field) is not None:
                    setattr(self._data_queue[-1], field, getattr(decoded_data, field))

        logger.debug(f"Callback size_dist from {self.serial_number}")

    def callback_read(self, characteristic: BleakGATTCharacteristic, data: bytearray) -> None:
        decoded_data = data.decode("utf-8").replace("\r", "").replace("\n", "")
        logger.info(f"Callback read from {self.serial_number}: {decoded_data}")

    @staticmethod
    def decode_std_chareristic(data: bytearray) -> Partector2DataStructure:
        decoded_data = Partector2DataStructure(
            ldsa=int.from_bytes(data[0:3], byteorder="little") / 100.0,
            particle_diameter=int.from_bytes(data[3:5], byteorder="little"),
            particle_number=int.from_bytes(data[5:8], byteorder="little"),
            temperature=int.from_bytes(data[8:9], byteorder="little"),
            relativ_humidity=int.from_bytes(data[9:10], byteorder="little"),
            device_status=int.from_bytes(data[10:12], byteorder="little")
            + (((int(data[19]) >> 1) & 0b01111111) << 16),
            battery_voltage=int.from_bytes(data[12:14], byteorder="little") / 100.0,
            particle_mass=int.from_bytes(data[14:18], byteorder="little") / 100.0,
        )

        return decoded_data

    @staticmethod
    def decode_aux_characteristic(data: bytearray) -> Partector2DataStructure:
        decoded_data = Partector2DataStructure(
            corona_voltage=int.from_bytes(data[0:2], byteorder="little"),
            diffusion_current=int.from_bytes(data[2:4], byteorder="little") / 100.0,
            deposition_voltage=int.from_bytes(data[4:6], byteorder="little"),
            flow_from_dp=int.from_bytes(data[6:8], byteorder="little"),
            ambient_pressure=int.from_bytes(data[8:10], byteorder="little"),
            em_amplitude1=int.from_bytes(data[10:12], byteorder="little"),
            em_amplitude2=int.from_bytes(data[12:14], byteorder="little"),
            em_gain1=int.from_bytes(data[14:16], byteorder="little"),
            em_gain2=int.from_bytes(data[16:18], byteorder="little"),
            diffusion_current_offset=int.from_bytes(data[18:20], byteorder="little"),
        )

        return decoded_data

    @staticmethod
    def decode_size_dist_characteristic(data: bytearray) -> Partector2DataStructure:
        decoded_data = Partector2DataStructure(
            dist_particle_number_10nm=int.from_bytes(data[0:3], byteorder="little"),
            dist_particle_number_16nm=int.from_bytes(data[3:6], byteorder="little"),
            dist_particle_number_26nm=int.from_bytes(data[6:9], byteorder="little"),
            dist_particle_number_43nm=int.from_bytes(data[9:12], byteorder="little"),
            dist_particle_number_70nm=int.from_bytes(data[12:15], byteorder="little"),
            dist_particle_number_114nm=int.from_bytes(data[15:18], byteorder="little"),
            dist_particle_number_185nm=int.from_bytes(data[18:21], byteorder="little"),
            dist_particle_number_300nm=int.from_bytes(data[21:24], byteorder="little"),
        )

        return decoded_data

    @classmethod
    def get_naneos_adv(cls, data: AdvertisementData) -> tuple[Optional[bytes], Optional[int]]:
        """
        Returns the custom advertisement data from Naneos devices.

        We are violating the BLE standard here by using the manufacturer data field for our own purposes.
        """
        # Because of the 22 byte limit we also encoded data into the manufacturer name field.
        naneos_adv = next(iter(data.manufacturer_data.keys())).to_bytes(2, byteorder="little")

        # security check 1: First byte must be "X"
        if naneos_adv[0].to_bytes(1).decode("utf-8") != "X":
            return (None, None)

        # add the data part of the manufacturer data
        naneos_adv += next(iter(data.manufacturer_data.values()))

        # security check 2: The data must be 22 bytes long
        if len(naneos_adv) != 22:
            return (None, None)

        serial_number = cls.get_serial_number(naneos_adv)

        return (naneos_adv, serial_number)

    @staticmethod
    def get_serial_number(data: bytes) -> int:
        return int.from_bytes(data[15:17], byteorder="little")



import base64
import datetime
from threading import Thread
from typing import Callable, Optional

import pandas as pd
import requests

from naneos.logger import get_naneos_logger
from naneos.protobuf import create_combined_entry, create_proto_p2_pro_garage

logger = get_naneos_logger(__name__)


class Partector2ProGarageUpload(Thread):
    URL = "https://hg3zkburji.execute-api.eu-central-1.amazonaws.com/dev/proto/v1"
    HEADERS = {"Content-Type": "application/json", "Accept": "application/json"}

    def __init__(
        self, df: pd.DataFrame, serial_number: int, callback: Optional[Callable[[bool], None]]
    ) -> None:
        super().__init__()
        self.df = df
        self.serial_number = serial_number
        self._callback = callback

    def run(self) -> None:
        try:
            ret = self.upload(self.df, self.serial_number)

            if self._callback:
                if ret.status_code == 200:
                    self._callback(True)
                else:
                    self._callback(False)
        except Exception as e:
            logger.error(f"Error in upload: {e}")
            if self._callback:
                self._callback(True)  # delete data because it was corrupted

    @staticmethod
    def get_body(upload_string: str) -> str:
        return f"""
            {{
                "gateway": "python_webhook",
                "data": "{upload_string}",
                "published_at": "{datetime.datetime.now().isoformat()}"
            }}
            """

    # can be used directly then its not threaded
    @classmethod
    def upload(cls, df: pd.DataFrame, serial_number: int) -> requests.Response:
        abs_time = int(datetime.datetime.now().timestamp())
        device = create_proto_p2_pro_garage(serial_number, abs_time, df)
        combined_entry = create_combined_entry(devices=[device], abs_timestamp=abs_time)

        proto_str = combined_entry.SerializeToString()
        # .decode() converts to str
        proto_str_base64 = base64.b64encode(proto_str).decode()

        body = cls.get_body(proto_str_base64)
        r = requests.post(cls.URL, headers=cls.HEADERS, data=body, timeout=10)
        # print(f"Status code: {r.status_code} text={r.text}")
        return r


if __name__ == "__main__":
    df = pd.read_pickle("/Users/huegi/gitlocal/naneos/naneos-devices/tests/df_garagae.pkl")

    abs_time = int(datetime.datetime.now().timestamp())
    serial_number = 777
    Partector2ProGarageUpload.upload(df, serial_number)



from dataclasses import dataclass
import datetime
from typing import Optional, Union


@dataclass
class Partector2DataStructure:
    # mandatory
    unix_timestamp: Optional[int] = None

    # optional
    runtime_min: Optional[float] = None
    device_status: Optional[int] = None
    ldsa: Optional[float] = None
    particle_number: Optional[int] = None
    particle_diameter: Optional[float] = None
    particle_mass: Optional[float] = None
    particle_surface: Optional[float] = None
    diffusion_current: Optional[float] = None
    diffusion_current_offset: Optional[float] = None
    corona_voltage: Optional[int] = None
    hires_adc1: Optional[float] = None
    hires_adc2: Optional[float] = None
    em_amplitude1: Optional[float] = None
    em_amplitude2: Optional[float] = None
    em_gain1: Optional[int] = None
    em_gain2: Optional[int] = None
    temperature: Optional[float] = None
    relativ_humidity: Optional[int] = None
    deposition_voltage: Optional[int] = None
    battery_voltage: Optional[float] = None
    flow_from_dp: Optional[float] = None
    differential_pressure: Optional[int] = None
    ambient_pressure: Optional[float] = None
    sigma: Optional[float] = None
    pump_current: Optional[float] = None
    pump_pwm: Optional[int] = None
    dist_steps: Optional[int] = None
    dist_particle_number_10nm: Optional[int] = None
    dist_particle_number_16nm: Optional[int] = None
    dist_particle_number_26nm: Optional[int] = None
    dist_particle_number_43nm: Optional[int] = None
    dist_particle_number_70nm: Optional[int] = None
    dist_particle_number_114nm: Optional[int] = None
    dist_particle_number_185nm: Optional[int] = None
    dist_particle_number_300nm: Optional[int] = None
    dist_current_0: Optional[float] = None
    dist_current_1: Optional[float] = None
    dist_current_2: Optional[float] = None
    dist_current_3: Optional[float] = None
    dist_current_4: Optional[float] = None
    cs_status: Optional[int] = None


PARTECTOR1_DATA_STRUCTURE_V_LEGACY: dict[str, Union[type[int], type[float]]] = {
    "unix_timestamp": int,
    "runtime_min": float,
    "batt_voltage": float,
    "idiff_global": float,
    "ucor_global": float,
    "EM": float,
    "DAC": float,
    "HVon": int,
    "idiffset": float,
    "flow_from_dp": float,
    "LDSA": float,
    "T": float,
    "RHcorr": float,
    "device_status": int,
    # "phase_angle": float,
}

# v298 and smaller
PARTECTOR2_DATA_STRUCTURE_V_LEGACY: dict[str, Union[type[int], type[float]]] = {
    "unix_timestamp": int,
    "runtime_min": float,
    "idiff_global": float,
    "ucor_global": int,
    "hiresADC1": float,
    "hiresADC2": float,
    "EM_amplitude1": float,
    "EM_amplitude2": float,
    "T": float,
    "RHcorr": int,
    "device_status": int,
    "deposition_voltage": int,
    "batt_voltage": float,
    "flow_from_dp": float,
    "LDSA": float,
    "diameter": float,
    "number": int,
    "dP": int,
    "P_average": float,
}

# TODO: wait for martin to implement this
PARTECTOR2_DATA_STRUCTURE_V299: dict[str, Union[type[int], type[float]]] = {
    "unix_timestamp": int,
    "runtime_min": float,
    "idiff_global": float,
    "ucor_global": int,
    "hiresADC1": float,
    "hiresADC2": float,
    "EM_amplitude1": float,
    "EM_amplitude2": float,
    "T": float,
    "RHcorr": int,
    "device_status": int,
    "deposition_voltage": int,
    "batt_voltage": float,
    "flow_from_dp": float,
    "LDSA": float,
    "diameter": float,
    "number": int,
    "dP": int,
    "P_average": float,
    "em_gain1": float,
    "em_gain2": float,
}

PARTECTOR2_PRO_DATA_STRUCTURE_V311: dict[str, Union[type[int], type[float]]] = {
    "unix_timestamp": int,
    "runtime_min": float,
    "number": int,
    "diameter": float,
    "LDSA": float,
    "surface": float,  # not existing in protobuf
    "particle_mass": float,
    "sigma": float,  # not existing in protobuf
    "idiff_global": float,
    "ucor_global": int,
    "deposition_voltage": int,
    "T": float,
    "RHcorr": int,
    "P_average": float,
    "flow_from_dp": float,
    "batt_voltage": float,
    "pump_current": float,  # not existing in protobuf
    "device_status": int,
    "pump_pwm": int,  # not existing in protobuf
    "steps": int,  # not existing in protobuf
    "particle_number_10nm": int,
    "particle_number_16nm": int,
    "particle_number_26nm": int,
    "particle_number_43nm": int,
    "particle_number_70nm": int,
    "particle_number_114nm": int,
    "particle_number_185nm": int,
    "particle_number_300nm": int,
    "current_0": float,  # not existing in protobuf
    "current_1": float,  # not existing in protobuf
    "current_2": float,  # not existing in protobuf
    "current_3": float,  # not existing in protobuf
    "current_4": float,  # not existing in protobuf
    "em_gain1": float,
    "em_gain2": float,
}

PARTECTOR2_PRO_CS_DATA_STRUCTURE_V315: dict[str, Union[type[int], type[float]]] = {
    "unix_timestamp": int,
    "runtime_min": float,
    "number": int,
    "diameter": float,
    "LDSA": float,
    "surface": float,  # not existing in protobuf
    "particle_mass": float,
    "sigma": float,  # not existing in protobuf
    "idiff_global": float,
    "ucor_global": int,
    "deposition_voltage": int,
    "T": float,
    "RHcorr": int,
    "P_average": float,
    "flow_from_dp": float,
    "batt_voltage": float,
    "pump_current": float,  # not existing in protobuf
    "device_status": int,
    "pump_pwm": int,  # not existing in protobuf
    "steps": int,  # not existing in protobuf
    "particle_number_10nm": int,
    "particle_number_16nm": int,
    "particle_number_26nm": int,
    "particle_number_43nm": int,
    "particle_number_70nm": int,
    "particle_number_114nm": int,
    "particle_number_185nm": int,
    "particle_number_300nm": int,
    "current_0": float,  # not existing in protobuf
    "current_1": float,  # not existing in protobuf
    "current_2": float,  # not existing in protobuf
    "current_3": float,  # not existing in protobuf
    "current_4": float,  # not existing in protobuf
    "em_gain1": float,
    "em_gain2": float,
    "cs_status": int,
}



from naneos.protobuf.protobuf import (
    create_combined_entry,
    create_proto_p1,
    create_proto_p2,
    create_proto_p2_pro,
    create_proto_p2_pro_garage,
)

__all__ = [
    "create_combined_entry",
    "create_proto_p1",
    "create_proto_p2",
    "create_proto_p2_pro",
    "create_proto_p2_pro_garage",
]



from typing import Optional

from naneos.partector.blueprints._data_structure import PARTECTOR2_PRO_DATA_STRUCTURE_V311
from naneos.partector.blueprints._partector_blueprint import PartectorBluePrint


class Partector2Pro(PartectorBluePrint):
    def __init__(
        self,
        serial_number: Optional[int] = None,
        port: Optional[str] = None,
        verb_freq: int = 1,
        hw_version: str = "P2_pro",
    ) -> None:
        super().__init__(serial_number, port, verb_freq, hw_version)

    def _init_serial_data_structure(self) -> None:
        self._data_structure = PARTECTOR2_PRO_DATA_STRUCTURE_V311

    def _set_verbose_freq(self, freq: int) -> None:
        if freq == 0:
            self._write_line("X0000!")
        else:
            self._write_line("h2001!")  # activates harmonics output
            self._write_line("M0004!")  # activates size dist mode
            self._write_line("X0006!")  # activates verbose mode


if __name__ == "__main__":
    import time

    from naneos.partector import scan_for_serial_partectors

    partectors = scan_for_serial_partectors()
    partectors = partectors["P2_Pro"]

    if not partectors:
        print("No Partector found!")
        exit()

    serial_number = next(iter(partectors.keys()))

    p2 = Partector2Pro(serial_number=serial_number)

    # print(p2.write_line("M?", 1))
    time.sleep(5)
    df = p2.get_data_pandas()
    print(df)
    # df.to_pickle("/Users/huegi/gitlocal/naneos/naneos-devices/tests/p2_pro_test_data.pkl")
    p2.close()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: protoV1.proto
# Protobuf Python Version: 4.25.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\rprotoV1.proto"\xb0\x01\n\x0c\x43ombinedData\x12\x15\n\rabs_timestamp\x18\x01 \x01(\r\x12\x18\n\x07\x64\x65vices\x18\x02 \x03(\x0b\x32\x07.Device\x12%\n\x0egateway_points\x18\x03 \x03(\x0b\x32\r.GatewayPoint\x12\'\n\x0fposition_points\x18\x04 \x03(\x0b\x32\x0e.PositionPoint\x12\x1f\n\x0bwind_points\x18\x05 \x03(\x0b\x32\n.WindPoint"`\n\x06\x44\x65vice\x12\x11\n\x04type\x18\x01 \x01(\rH\x00\x88\x01\x01\x12\x15\n\rserial_number\x18\x02 \x01(\r\x12#\n\rdevice_points\x18\x03 \x03(\x0b\x32\x0c.DevicePointB\x07\n\x05_type"\xa2\x08\n\x0b\x44\x65vicePoint\x12\x16\n\ttimestamp\x18\x01 \x01(\rH\x00\x88\x01\x01\x12\x1a\n\rdevice_status\x18\x02 \x01(\rH\x01\x88\x01\x01\x12\x0c\n\x04ldsa\x18\x03 \x01(\r\x12!\n\x19\x61verage_particle_diameter\x18\x04 \x01(\r\x12%\n\x1dparticle_number_concentration\x18\x05 \x01(\r\x12\x13\n\x0btemperature\x18\x06 \x01(\x05\x12\x19\n\x11relative_humidity\x18\x07 \x01(\r\x12\x17\n\x0f\x62\x61ttery_voltage\x18\x08 \x01(\r\x12\x15\n\rparticle_mass\x18\t \x01(\r\x12\x16\n\x0e\x63orona_voltage\x18\n \x01(\r\x12\x19\n\x11\x64iffusion_current\x18\x0b \x01(\r\x12\x1a\n\x12\x64\x65position_voltage\x18\x0c \x01(\r\x12\x0c\n\x04\x66low\x18\r \x01(\r\x12\x18\n\x10\x61mbient_pressure\x18\x0e \x01(\r\x12\x1b\n\x13\x65lectrometer_offset\x18\x0f \x01(\r\x12\x1d\n\x15\x65lectrometer_2_offset\x18\x10 \x01(\r\x12\x19\n\x11\x65lectrometer_gain\x18\x11 \x01(\r\x12\x1b\n\x13\x65lectrometer_2_gain\x18\x12 \x01(\r\x12 \n\x18\x64iffusion_current_offset\x18\x13 \x01(\r\x12\x1c\n\x14particle_number_10nm\x18\x14 \x01(\r\x12\x1c\n\x14particle_number_16nm\x18\x15 \x01(\r\x12\x1c\n\x14particle_number_26nm\x18\x16 \x01(\r\x12\x1c\n\x14particle_number_43nm\x18\x17 \x01(\r\x12\x1c\n\x14particle_number_70nm\x18\x18 \x01(\r\x12\x1d\n\x15particle_number_114nm\x18\x19 \x01(\r\x12\x1d\n\x15particle_number_185nm\x18\x1a \x01(\r\x12\x1d\n\x15particle_number_300nm\x18\x1b \x01(\r\x12\x0f\n\x07surface\x18\x1c \x01(\r\x12\x17\n\x0fsigma_size_dist\x18\x1d \x01(\r\x12\x17\n\x0fsteps_inversion\x18\x1e \x01(\r\x12\x16\n\x0e\x63urrent_dist_0\x18\x1f \x01(\r\x12\x16\n\x0e\x63urrent_dist_1\x18  \x01(\r\x12\x16\n\x0e\x63urrent_dist_2\x18! \x01(\r\x12\x16\n\x0e\x63urrent_dist_3\x18" \x01(\r\x12\x16\n\x0e\x63urrent_dist_4\x18# \x01(\r\x12\x14\n\x0cpump_current\x18$ \x01(\r\x12\x10\n\x08pump_pwm\x18% \x01(\r\x12\x16\n\tcs_status\x18& \x01(\x05H\x02\x88\x01\x01\x42\x0c\n\n_timestampB\x10\n\x0e_device_statusB\x0c\n\n_cs_status"\xb4\x02\n\x0cGatewayPoint\x12\x16\n\ttimestamp\x18\x01 \x01(\rH\x00\x88\x01\x01\x12\x15\n\rserial_number\x18\x02 \x01(\r\x12\x18\n\x10\x66irmware_version\x18\x03 \x01(\r\x12\x13\n\x0b\x66ree_memory\x18\x04 \x01(\r\x12\x11\n\tfree_heap\x18\x05 \x01(\r\x12\x1f\n\x17largest_free_block_heap\x18\x06 \x01(\r\x12\x17\n\x0f\x63\x65llular_signal\x18\x07 \x01(\x05\x12\x17\n\x0f\x62\x61ttery_int_soc\x18\x08 \x01(\r\x12\x17\n\x0f\x62\x61ttery_ext_soc\x18\t \x01(\r\x12\x1b\n\x13\x62\x61ttery_ext_voltage\x18\n \x01(\r\x12\x1c\n\x14\x63harging_ext_voltage\x18\x0b \x01(\rB\x0c\n\n_timestamp"Z\n\rPositionPoint\x12\x16\n\ttimestamp\x18\x01 \x01(\rH\x00\x88\x01\x01\x12\x10\n\x08latitude\x18\x02 \x01(\x02\x12\x11\n\tlongitude\x18\x03 \x01(\x02\x42\x0c\n\n_timestamp"\x81\x01\n\tWindPoint\x12\x16\n\ttimestamp\x18\x01 \x01(\rH\x00\x88\x01\x01\x12\x17\n\nwind_speed\x18\x02 \x01(\rH\x01\x88\x01\x01\x12\x17\n\nwind_angle\x18\x03 \x01(\rH\x02\x88\x01\x01\x42\x0c\n\n_timestampB\r\n\x0b_wind_speedB\r\n\x0b_wind_angleb\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "protoV1_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _globals["_COMBINEDDATA"]._serialized_start = 18
    _globals["_COMBINEDDATA"]._serialized_end = 194
    _globals["_DEVICE"]._serialized_start = 196
    _globals["_DEVICE"]._serialized_end = 292
    _globals["_DEVICEPOINT"]._serialized_start = 295
    _globals["_DEVICEPOINT"]._serialized_end = 1353
    _globals["_GATEWAYPOINT"]._serialized_start = 1356
    _globals["_GATEWAYPOINT"]._serialized_end = 1664
    _globals["_POSITIONPOINT"]._serialized_start = 1666
    _globals["_POSITIONPOINT"]._serialized_end = 1756
    _globals["_WINDPOINT"]._serialized_start = 1759
    _globals["_WINDPOINT"]._serialized_end = 1888
# @@protoc_insertion_point(module_scope)



from typing import Optional

from naneos.logger import get_naneos_logger
from naneos.partector.blueprints._data_structure import (
    PARTECTOR2_DATA_STRUCTURE_V299,
    PARTECTOR2_DATA_STRUCTURE_V_LEGACY,
)
from naneos.partector.blueprints._partector_blueprint import PartectorBluePrint

logger = get_naneos_logger(__name__)


class Partector2(PartectorBluePrint):
    def __init__(
        self, serial_number: Optional[int] = None, port: Optional[str] = None, verb_freq: int = 1
    ) -> None:
        super().__init__(serial_number, port, verb_freq, "P2")

    def _init_serial_data_structure(self) -> None:
        if self._fw > 298:
            self._data_structure = PARTECTOR2_DATA_STRUCTURE_V299
            logger.info(f"SN{self._sn} has FW{self._fw}. -> Using V299 data structure.")
        else:
            self._data_structure = PARTECTOR2_DATA_STRUCTURE_V_LEGACY
            logger.info(f"SN{self._sn} has FW{self._fw}. -> Using legacy data structure.")

    def _set_verbose_freq(self, freq: int) -> None:
        """
        Set the frequency of the verbose output.

        :param int freq: Frequency of the verbose output in Hz. (0: off, 1: 1Hz, 2: 10Hz, 3: 100Hz)
        """

        if freq < 0 or freq > 3:
            raise ValueError("Frequency must be between 0 and 3!")

        self._write_line(f"X000{freq}!")


if __name__ == "__main__":
    import time

    from naneos.partector import scan_for_serial_partectors

    # production scanning:
    p2 = Partector2(serial_number=8112)
    time.sleep(5)
    print(p2.get_data_pandas())
    p2.close()
